/**************************************************************************
 *
 * This file is part of Buildbot.  Buildbot is free software: you can
 * redistribute it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <sys/errno.h>

#include "tdsrom-cmds.h"
#include "ib.h"

#define MIN(_a, _b) ((_a) > (_b) ? (_b) : (_a))

#define ARG_OP_PI 0x10000000
#define ARG_OP_ID 0x20000000
#define ARG_OP_PG 0x40000000
#define ARG_OP_ER 0x80000000

int abort_requested = 0;
int debug = 0;
int Dev;

static struct option long_options[] = {
	{ "read", required_argument, 0, 'r' },
	{ "write", required_argument, 0, 'w' },
	{ "program", required_argument, 0, 'p' },
	{ "base", required_argument, 0, 'b' },
	{ "length", required_argument, 0, 'l' },
	{ "skip", required_argument, 0, 's' },
	{ "debug", no_argument, 0, 'd' },
	{ "exec", optional_argument, 0, 'x' },
	{ "help", no_argument, 0, 'h' },
	{ NULL, 0, 0, 0 }
};

static void usage(void)
{
	fprintf(stderr, "usage:\n"
		"--read         -r <filename>  read from memory to file\n"
		"--write        -w <filename>  read from file to memory\n"
		"--program      -p <filename>  read from file to flash\n"
		"--base         -b <base>      base address for read/write/program\n"
		"--length       -l <length>    length of data to be read or written\n"
		"--debug        -d             enable debug logging\n"
		"--exec         -x <addr>      execute starting at address\n"
		"--skip         -s <addr>      skip programming until address\n"
		"");
}

static uint32_t to_number(char *s)
{
	uint32_t val;
	char *endp;

	if (*s == '0' && *(s+1) == 'x') {
		val = strtoul(s+2, &endp, 16);
		if (*endp != '\0') {
			fprintf(stderr, "failed to parse: [%s]\n", s);
			return 0;
		}
	} else {
		val = strtoul(s, &endp, 10);
		if (*endp != '\0') {
			fprintf(stderr, "failed to parse: [%s]\n", s);
			return 0;
		}
	}
	return val;
}

int main(int argc, char **argv)
{
	uint32_t len = 0, addr = 0, base = 0, length = 0, exec_addr = 0, skip_addr = 0, stat;
	char c;
	uint8_t buf[1024];
	int optidx;
	FILE *file = NULL;
	int exec_op = 0, read_op = 0, write_op = 0, identify_flash_op = 0, erase_flash_op = 0, flash_write_op = 0;
	int readlen;
	time_t start;
	int i, j;

	while((c = getopt_long(argc, argv, "r:w:p:b:l:s:x:hd",
			       long_options, &optidx)) != -1) {
		switch(c) {
		case 'h':
			usage();
			return 0;
		case 'l':
			if (length) {
				fprintf(stderr, "length given twice");
				return 1;
			}
			length = to_number(optarg);
			break;
		case 's':
			skip_addr = to_number(optarg);
			break;
		case 'b':
			if (base) {
				fprintf(stderr, "base given twice");
				return 1;
			}
			base = to_number(optarg);
			break;
		case 'r':
			if (file) {
				fprintf(stderr, "read given twice");
				return 1;
			}
			file = fopen(optarg, "wb");
			if (!file) {
				fprintf(stderr, "failed to open output file: %s\n", strerror(errno));
				return 1;

			}
			read_op = 1;
			break;
		case 'x':
			exec_op = 1;
			if (write_op) exec_addr = base;
			if (optarg) exec_addr = to_number(optarg);
			if (!exec_addr) {
				fprintf(stderr, "exec needs an addr or a write");
				return 1;
			}
			break;
		case 'w':
			if (file) {
				fprintf(stderr, "read given twice");
				return 1;
			}
			file = fopen(optarg, "rb");
			if (!file) {
				fprintf(stderr, "failed to open input file: %s\n", strerror(errno));
				return 1;

			}

			write_op = 1;
			break;
		case 'p':
			if (file) {
				fprintf(stderr, "read given twice");
				return 1;
			}
			file = fopen(optarg, "rb");
			if (!file) {
				fprintf(stderr, "failed to open input file: %s\n", strerror(errno));
				return 1;

			}
			flash_write_op = 1;
			break;

		case 'i':
			identify_flash_op = 1;
			break;
		case 'e':
			erase_flash_op = 1;
			break;
		case 'd':
			debug++;
			break;
		}
	}

	// signal(SIGINT, sigint_handler);

	Dev = ibdev(0, 29, 0, 20, 1, 0);
	if (ibsta & ERR) {
		printf("Unable to open device\nibsta = 0x%x iberr = %d\n",
		       ibsta, iberr);
		return 1;
	}

	ibclr (Dev);
	//if (ibsta & ERR) {
	//	GPIBCleanup(Dev, "Unable to clear device");
	//	return 1;
	//}

	time(&start);
	for(addr = base; addr < base + length && !abort_requested;) {
		len = MIN(read_op?512:128, base + length - addr);
		if (read_op) {
			if (debug) printf("Read: 0x%.8x len %d\n", addr, len);
			if (read_memory(addr, buf, len) == -1)
				return 1;

			if (fwrite(buf, 1, len, file) != len) {
				fprintf(stderr, "short write\n");
				return 1;
			}
			addr += len;
		} else if (write_op) {
			if (debug) printf("Write: 0x%.8x len %d\n", addr, len);
			readlen = fread(buf, 1, len, file);
			if (readlen == 0)
				break;

			if (readlen == -1) {
				fprintf(stderr, "read: %s\n", strerror(errno));
				return 1;
			}
			if ((addr % 0x1000) == 0)
				fprintf(stderr, "WRITE %08x %3d%%\r",  addr, ((addr -base) * 100)/length);
			if (write_memory(addr, buf, readlen) == -1)
				return 1;
			addr += readlen;
		} else if (flash_write_op) {
			if (debug) printf("Program: 0x%.8x len %d\n", addr, len);

			readlen = fread(buf, 1, len, file);
			if (readlen == 0) break;
			if (readlen == -1) {
				fprintf(stderr, "read: %s\n", strerror(errno));
				return 1;
			}

			if (skip_addr > addr) { addr += len; continue; }
			if ((addr % 0x40000) == 0) {
				printf("Erase block 0x%.8x\n", addr);
				j = execute(0x05010000, ARG_OP_ER | addr, &stat, 4);
				if (stat != 0x80808080) printf("Didn't erase! stat %.8x\n", stat);

				if (debug) { for (i=0; i<j; i++) printf(" %.2X", 0xff & buf[i]); printf("  :  "); }
				for (i=0; i<j; i++) printf("%c", isprint(buf[i]) ? buf[i] : '.');
				printf("\n");
			}
			for (i=0; i<readlen; i++) if (buf[i] != 0xff) break;
			if (i == readlen) {
				addr += readlen;
				continue; // no bytes with 0s to program
			}

			if ((addr % 0x1000) == 0) fprintf(stderr, "PROGRAM %08x %3d%%\r",  addr, ((addr -base) * 100)/length);
			j = execute(0x05010000, ARG_OP_PG | addr, buf, readlen);

			if (debug) for (i=0; i<j; i++) printf("%c", isprint(buf[i]) ? buf[i] : '.');
			if (debug) printf("\n");

			addr += readlen;
		}
	}
	fclose(file);

	if (exec_op) {
		if (debug) printf("exec addr %.8x\n", exec_addr);
		for (i=0; i<20; i++) buf[i] = 0x0;

		readlen = execute(exec_addr, ARG_OP_PI, buf, 20);

		if (debug) { for (i=0; i<readlen; i++) printf(" %.2X", 0xff & buf[i]); printf("  :  "); }
		for (i=0; i<readlen; i++) printf("%c", isprint(buf[i]) ? buf[i] : '.');
		printf("\n");

		if (debug) printf("exec addr %.8x\n", exec_addr);
		for (i=0; i<20; i++) buf[i] = 0x0;

		readlen = execute(exec_addr, ARG_OP_ID | 0x01000000, buf, 20);

		printf("FLASH mf mf mf mf id id id id st st st st\n     ");
		for (i=0; i<readlen; i++) printf(" %.2X", 0xff & buf[i]);
		printf("\n");
#if 0
		/* Program code */
		if (debug) printf("exec addr %.8x\n", exec_addr);
		for (i=0; i<20; i++) buf[i] = i;

		readlen = execute(exec_addr, ARG_OP_PG | 0x013fff00, buf, 20);

		if (debug) { for (i=0; i<readlen; i++) printf(" %.2X", 0xff & buf[i]); printf("  :  "); }
		for (i=0; i<readlen; i++) printf("%c", isprint(buf[i]) ? buf[i] : '.');
		printf("\n");

		/* Erase code */
		if (debug) printf("exec addr %.8x\n", exec_addr);
		for (i=0; i<20; i++) buf[i] = i;

		readlen = execute(exec_addr, ARG_OP_ER | 0x013fff00, buf, 20);

		if (debug) { for (i=0; i<readlen; i++) printf(" %.2X", 0xff & buf[i]); printf("  :  "); }
		for (i=0; i<readlen; i++) printf("%c", isprint(buf[i]) ? buf[i] : '.');
		printf("\n");
#endif
	}

	ibonl(Dev, 0);
	return 0;
}
